import logging
from abc import ABC
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from types import ModuleType
from typing import Any, Dict, Optional, Union

import yaml
from curies import Converter
from linkml_runtime import SchemaView
from linkml_runtime.processing.referencevalidator import ReferenceValidator
from linkml_runtime.utils.introspection import package_schemaview
from linkml_runtime.utils.yamlutils import YAMLRoot
from pydantic import BaseModel

from linkml_transformer.datamodel.transformer_model import (
    ClassDerivation,
    CollectionType,
    SlotDerivation,
    TransformationSpecification,
)
from linkml_transformer.transformer.inference import induce_missing_values

logger = logging.getLogger(__name__)


OBJECT_TYPE = Union[Dict[str, Any], BaseModel, YAMLRoot]
"""An object can be a plain python dict, a pydantic object, or a linkml YAMLRoot"""


@dataclass
class Transformer(ABC):
    """
    Base class for all transformers.

    A transformer will generate an instance of a target class from
    an instance of a source class, making use of a specification.

    This is an abstract class. Different implementations will
    subclass this
    """

    source_schemaview: SchemaView = None
    """A view over the schema describing the input/source object."""

    specification: TransformationSpecification = None
    """A specification of how to generate target objects from source objects."""

    _derived_specification: TransformationSpecification = None
    """A specification with inferred missing values."""

    target_schemaview: Optional[SchemaView] = None
    """A view over the schema describing the output/target object."""

    target_module: Optional[ModuleType] = None
    """The python module which the target object should conform to."""

    prefix_map: Optional[Dict[str, str]] = None
    """Additional prefixes"""

    _curie_converter: Converter = None

    def transform(self, obj: OBJECT_TYPE, source_type: str = None) -> OBJECT_TYPE:
        """
        Transform source object into an instance of the target class.

        :param obj:
        :param source_type:
        :return:
        """
        raise NotImplementedError

    def load_source_schema(self, path: Union[str, Path]):
        """
        Sets source_schemaview from a schema path.

        :param path:
        :return:
        """
        self.source_schemaview = SchemaView(str(path))

    def load_transformer_specification(self, path: Union[str, Path]):
        """
        Sets specification from a schema path.

        :param path:
        :return:
        """
        # self.specification = yaml_loader.load(str(path), TransformationSpecification)
        with open(path) as f:
            obj = yaml.safe_load(f)
            # necessary to expand first
            normalizer = ReferenceValidator(
                package_schemaview("linkml_transformer.datamodel.transformer_model")
            )
            normalizer.expand_all = True
            obj = normalizer.normalize(obj)
            self.specification = TransformationSpecification(**obj)

    @property
    def derived_specification(self) -> Optional[TransformationSpecification]:
        if self._derived_specification is None:
            if self.specification is None:
                return None
            self._derived_specification = deepcopy(self.specification)
            induce_missing_values(self._derived_specification, self.source_schemaview)
        return self._derived_specification

    def _get_class_derivation(self, target_class_name) -> ClassDerivation:
        spec = self.derived_specification
        matching_tgt_class_derivs = [
            deriv
            for deriv in spec.class_derivations.values()
            if deriv.populated_from == target_class_name
        ]
        logger.debug(f"Target class derivs={matching_tgt_class_derivs}")
        if len(matching_tgt_class_derivs) != 1:
            raise ValueError(f"Could not find what to derive from a source {target_class_name}")
        return matching_tgt_class_derivs[0]

    def _coerce_to_multivalued(
        self, slot_derivation: SlotDerivation, class_derivation: ClassDerivation
    ):
        cast_as = slot_derivation.cast_collection_as
        if cast_as and cast_as == CollectionType.MultiValued:
            return True
        sv = self.target_schemaview
        if sv:
            slot = sv.induced_slot(slot_derivation.name, class_derivation.name)
            if slot.multivalued:
                return True
        return False

    def _coerce_to_singlevalued(
        self, slot_derivation: SlotDerivation, class_derivation: ClassDerivation
    ):
        cast_as = slot_derivation.cast_collection_as
        if cast_as and cast_as == CollectionType(CollectionType.SingleValued):
            return True
        sv = self.target_schemaview
        if sv:
            slot = sv.induced_slot(slot_derivation.name, class_derivation.name)
            if not slot.multivalued:
                return True
        return False

    @property
    def curie_converter(self) -> Converter:
        if not self._curie_converter:
            self._curie_converter = Converter([])
            for prefix in self.source_schemaview.schema.prefixes.values():
                self._curie_converter.add_prefix(prefix.prefix_prefix, prefix.prefix_reference)
            for prefix in self.specification.prefixes.values():
                self._curie_converter.add_prefix(prefix.key, prefix.value)
        return self._curie_converter

    def expand_curie(self, curie: str) -> str:
        return self.curie_converter.expand(curie)

    def compress_uri(self, uri: str) -> str:
        return self.curie_converter.compress(uri)



from linkml_runtime import SchemaView

from linkml_transformer.datamodel.transformer_model import TransformationSpecification


def induce_missing_values(
    specification: TransformationSpecification, source_schemaview: SchemaView
):
    """
    Infer missing values in a specification.

    Currently only uses copy directives.

    :param specification:
    :param source_schemaview:
    :return:
    """
    for cd in specification.class_derivations.values():
        if not cd.populated_from:
            cd.populated_from = cd.name
    for cd in specification.class_derivations.values():
        for sd in cd.slot_derivations.values():
            if sd.populated_from is None and sd.expr is None:
                sd.populated_from = sd.name
            if not sd.range:
                if sd.populated_from:
                    if cd.populated_from not in source_schemaview.all_classes():
                        continue
                    source_induced_slot = source_schemaview.induced_slot(
                        sd.populated_from, cd.populated_from
                    )
                    source_induced_slot_range = source_induced_slot.range
                    for range_cd in specification.class_derivations.values():
                        if range_cd.populated_from == source_induced_slot_range:
                            sd.range = range_cd.name



from linkml_transformer.datamodel.transformer_model import TransformationSpecification


def invert_transformation_specification(
    specification: TransformationSpecification,
) -> TransformationSpecification:
    raise NotImplementedError



import logging
from dataclasses import dataclass
from typing import Any, Dict, Optional, Type, Union

from linkml_runtime.index.object_index import ObjectIndex
from linkml_runtime.utils.eval_utils import eval_expr
from linkml_runtime.utils.yamlutils import YAMLRoot
from pydantic import BaseModel

from linkml_transformer.transformer.transformer import OBJECT_TYPE, Transformer
from linkml_transformer.utils.dynamic_object import dynamic_object

DICT_OBJ = Dict[str, Any]


logger = logging.getLogger(__name__)


@dataclass
class ObjectTransformer(Transformer):
    """
    A Transformer that works on in-memory dict objects.

    This works recursively
    """

    object_index: ObjectIndex = None

    def index(self, source_obj: Any, target: str = None):
        """
        Create an index over a container object.

        :param source_obj: source data structure to be indexed
        :param target: class to convert source object into
        """
        if isinstance(source_obj, dict):
            if target is None:
                [target] = [
                    c.name for c in self.source_schemaview.all_classes().values() if c.tree_root
                ]
            if target is None:
                raise ValueError(f"target must be passed if source_obj is dict: {source_obj}")
            source_obj_typed = dynamic_object(source_obj, self.source_schemaview, target)
            self.object_index = ObjectIndex(source_obj_typed, schemaview=self.source_schemaview)
        else:
            self.object_index = ObjectIndex(source_obj, schemaview=self.source_schemaview)

    def transform(
        self,
        source_obj: OBJECT_TYPE,
        source_type: str = None,
        target_type: str = None,
    ) -> Union[DICT_OBJ, Any]:
        """
        Transform a source object into a target object.

        :param source_obj:
        :param source_type:
        :return:
        """
        sv = self.source_schemaview
        if source_type is None:
            [source_type] = [c.name for c in sv.all_classes().values() if c.tree_root]
        if source_type in sv.all_types():
            if target_type:
                if target_type == "string":
                    return str(source_obj)
                elif target_type == "integer":
                    return int(source_obj)
                elif target_type == "float" or target_type == "double":
                    return float(source_obj)
                elif target_type == "uri":
                    return self.expand_curie(source_obj)
                elif target_type == "curie":
                    return self.compress_uri(source_obj)
            return source_obj
        if source_type in sv.all_enums():
            # TODO: enum derivations
            return str(source_obj)
        if isinstance(source_obj, (BaseModel, YAMLRoot)):
            source_obj_typed = source_obj
            source_obj = vars(source_obj)
        else:
            source_obj_typed = None
        if not isinstance(source_obj, dict):
            logger.warning(f"Unexpected: {source_obj} for type {source_type}")
            return source_obj
        class_deriv = self._get_class_derivation(source_type)
        tgt_attrs = {}
        for slot_derivation in class_deriv.slot_derivations.values():
            v = None
            source_class_slot = None
            if slot_derivation.populated_from:
                v = source_obj.get(slot_derivation.populated_from, None)
                source_class_slot = sv.induced_slot(slot_derivation.populated_from, source_type)
                logger.debug(
                    f"Pop slot {slot_derivation.name} => {v} using {slot_derivation.populated_from} // {source_obj}"
                )
            elif slot_derivation.expr:
                if self.object_index:
                    if not source_obj_typed:
                        source_obj_dyn = dynamic_object(source_obj, sv, source_type)
                    else:
                        source_obj_dyn = source_obj_typed
                    ctxt_obj = self.object_index.bless(source_obj_dyn)
                    ctxt_dict = {
                        k: getattr(ctxt_obj, k)
                        for k in ctxt_obj._attributes()
                        if not k.startswith("_")
                    }
                else:
                    do = dynamic_object(source_obj, sv, source_type)
                    ctxt_dict = vars(do)
                v = eval_expr(slot_derivation.expr, **ctxt_dict, NULL=None)
            else:
                source_class_slot = sv.induced_slot(slot_derivation.name, source_type)
                v = source_obj.get(slot_derivation.name, None)
            if source_class_slot and v is not None:
                target_range = slot_derivation.range
                source_class_slot_range = source_class_slot.range
                if source_class_slot.multivalued:
                    if isinstance(v, list):
                        v = [self.transform(v1, source_class_slot_range, target_range) for v1 in v]
                    elif isinstance(v, dict):
                        v = [self.transform(v1, source_class_slot_range, target_range) for v1 in v]
                    else:
                        v = [v]
                else:
                    v = self.transform(v, source_class_slot_range, target_range)
                if (
                    self._coerce_to_multivalued(slot_derivation, class_deriv)
                    and v is not None
                    and not isinstance(v, list)
                ):
                    v = [v]
                if self._coerce_to_singlevalued(slot_derivation, class_deriv) and isinstance(
                    v, list
                ):
                    if len(v) > 1:
                        raise ValueError(f"Cannot coerce multiple values {v}")
                    if len(v) == 0:
                        v = None
                    else:
                        v = v[0]
            tgt_attrs[str(slot_derivation.name)] = v
        return tgt_attrs

    def transform_object(
        self,
        source_obj: Union[YAMLRoot, BaseModel],
        target_class: Optional[Union[Type[YAMLRoot], Type[BaseModel]]] = None,
    ) -> Union[YAMLRoot, BaseModel]:
        typ = type(source_obj)
        typ_name = typ.__name__
        # if isinstance(source_obj, YAMLRoot):
        #    source_obj_dict = json_dumper.to_dict(source_obj)
        # elif isinstance(source_obj, BaseModel):
        #    source_obj_dict = source_obj.dict()
        # else:
        #    raise ValueError(f"Do not know how to handle type: {typ}")
        tr_obj_dict = self.transform(source_obj, typ_name)
        return target_class(**tr_obj_dict)



"""Command line interface for linkml-transformer."""
import logging
import sys

import click

__all__ = [
    "main",
]

import yaml
from linkml_runtime import SchemaView
from linkml_runtime.dumpers import yaml_dumper

from linkml_transformer.schema_mapper.schema_mapper import SchemaMapper
from linkml_transformer.transformer.object_transformer import ObjectTransformer

schema_option = click.option("-s", "--schema", help="Path to source schema.")
transformer_specification_option = click.option(
    "-T", "--transformer-specification", help="Path to transformer specification."
)
output_option_wb = click.option(
    "-o",
    "--output",
    type=click.File(mode="wb"),
    default=sys.stdout,
    help="Output file.",
)
output_option = click.option("-o", "--output", help="Output file.")
output_format_options = click.option(
    "-O",
    "--output-format",
    type=click.Choice(["json", "yaml", "pickle", "md"]),
    default="yaml",
    help="Output format.",
)


@click.group()
@click.option("-v", "--verbose", count=True)
@click.option("-q", "--quiet")
# @click.version_option(__version__)
def main(verbose: int, quiet: bool):
    """CLI for linkml-transformer."""
    logger = logging.getLogger()
    if verbose >= 2:
        logger.setLevel(level=logging.DEBUG)
    elif verbose == 1:
        logger.setLevel(level=logging.INFO)
    else:
        logger.setLevel(level=logging.WARNING)
    if quiet:
        logger.setLevel(level=logging.ERROR)
    logger.info(f"Logger {logger.name} set to level {logger.level}")


@main.command()
@output_option
@transformer_specification_option
@schema_option
@output_format_options
@click.option("--source-type")
@click.argument("input")
def map_data(
    input,
    schema,
    source_type,
    transformer_specification,
    output,
    output_format,
    **kwargs,
):
    """Map data in a source schema using a transformation."""
    logging.info(f"Transforming {input} conforming to {schema} using {transformer_specification}")
    tr = ObjectTransformer()
    tr.source_schemaview = SchemaView(schema)
    tr.load_transformer_specification(transformer_specification)
    with open(input) as file:
        input_obj = yaml.safe_load(file)
    tr.index(input_obj, source_type)
    tr_obj = tr.transform(input_obj, source_type)
    if output:
        outfile = open(output, "w", encoding="utf-8")
    else:
        outfile = sys.stdout
    outfile.write(yaml_dumper.dumps(tr_obj))


@main.command()
@output_option
@transformer_specification_option
@output_format_options
@click.argument("schema")
def derive_schema(schema, transformer_specification, output, output_format, **kwargs):
    """Derive a schema from a source schema and a transformation specification.

    This can be thought of as "copying" the source to a target, using the transformation
    specification as a "patch"

    Notes:

        the implementation is currently incomplete; the derived schema may not be valid
        linkml, e.g. there may be "dangling" references.

    Example:

        linkml-tr derive-schema -T transform/personinfo-to-agent.transform.yaml source/personinfo.yaml
    """
    logging.info(f"Transforming {schema} using {transformer_specification}")
    tr = ObjectTransformer()
    tr.load_transformer_specification(transformer_specification)
    mapper = SchemaMapper(transformer=tr)
    mapper.source_schemaview = SchemaView(schema)
    target_schema = mapper.derive_schema()
    if output:
        outfile = open(output, "w", encoding="utf-8")
    else:
        outfile = sys.stdout
    outfile.write(yaml_dumper.dumps(target_schema))


if __name__ == "__main__":
    main()



from __future__ import annotations

from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, List, Literal, Optional, Union

from linkml_runtime.linkml_model import Decimal
from pydantic import BaseModel as BaseModel
from pydantic import Field

metamodel_version = "None"
version = "None"


class WeakRefShimBaseModel(BaseModel):
    __slots__ = "__weakref__"


class ConfiguredBaseModel(
    WeakRefShimBaseModel,
    validate_assignment=True,
    validate_all=True,
    underscore_attrs_are_private=True,
    extra="forbid",
    arbitrary_types_allowed=True,
):
    pass


class CollectionType(str, Enum):
    SingleValued = "SingleValued"
    MultiValued = "MultiValued"
    MultiValuedList = "MultiValuedList"
    MultiValuedDict = "MultiValuedDict"


class TransformationSpecification(ConfiguredBaseModel):
    """
    A collection of mappings between source and target classes
    """

    id: Optional[str] = Field(
        None, description="""Unique identifier for this transformation specification"""
    )
    title: Optional[str] = Field(
        None,
        description="""human readable title for this transformation specification""",
    )
    prefixes: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict, description="""maps prefixes to URL expansions"""
    )
    source_schema: Optional[str] = Field(
        None,
        description="""name of the schema that describes the source (input) objects""",
    )
    target_schema: Optional[str] = Field(
        None,
        description="""name of the schema that describes the target (output) objects""",
    )
    class_derivations: Optional[Dict[str, ClassDerivation]] = Field(
        default_factory=dict,
        description="""Instructions on how to derive a set of classes in the target schema from classes in the source schema.""",
    )
    enum_derivations: Optional[Dict[str, EnumDerivation]] = Field(
        default_factory=dict,
        description="""Instructions on how to derive a set of enums in the target schema""",
    )
    slot_derivations: Optional[Dict[str, SlotDerivation]] = Field(
        default_factory=dict,
        description="""Instructions on how to derive a set of top level slots in the target schema""",
    )


class ElementDerivation(ConfiguredBaseModel):
    """
    An abstract grouping for classes that provide a specification of how to  derive a target element from a source element.
    """

    name: str = Field(None, description="""Name of the element in the target schema""")
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[Dict[str, ElementDerivation]] = Field(default_factory=dict)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys are expressions""",
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )


class ClassDerivation(ElementDerivation):
    """
    A specification of how to derive a target class from a source class.
    """

    populated_from: Optional[str] = Field(
        None, description="""Name of the class in the source schema"""
    )
    joins: Optional[Dict[str, AliasedClass]] = Field(
        default_factory=dict,
        description="""Additional classes to be joined to derive instances of the target class""",
    )
    slot_derivations: Optional[Dict[str, SlotDerivation]] = Field(default_factory=dict)
    name: str = Field(None, description="""Name of the element in the target schema""")
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[Dict[str, ElementDerivation]] = Field(default_factory=dict)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys are expressions""",
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )


class AliasedClass(ConfiguredBaseModel):
    """
    alias-class key value pairs for classes
    """

    alias: str = Field(None, description="""name of the class to be aliased""")
    class_named: Optional[str] = Field(None, description="""local alias for the class""")


class SlotDerivation(ElementDerivation):
    """
    A specification of how to derive the value of a target slot from a source slot
    """

    name: str = Field(None, description="""Target slot name""")
    populated_from: Optional[str] = Field(None, description="""Source slot name""")
    expr: Optional[str] = Field(
        None,
        description="""An expression to be evaluated on the source object to derive the target slot. Should be specified using the LinkML expression language.""",
    )
    range: Optional[str] = Field(None)
    inverse_of: Optional[Inverse] = Field(
        None,
        description="""Used to specify a class-slot tuple that is the inverse of the derived/target slot. This is used primarily for mapping to relational databases or formalisms that do not allow multiple values. The class representing the repeated element has a foreign key slot inserted in that 'back references' the original multivalued slot.""",
    )
    hide: Optional[bool] = Field(None, description="""True if this is suppressed""")
    type_designator: Optional[bool] = Field(None)
    cast_collection_as: Optional[CollectionType] = Field(None)
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[Dict[str, ElementDerivation]] = Field(default_factory=dict)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys are expressions""",
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )


class EnumDerivation(ElementDerivation):
    """
    A specification of how to derive the value of a target enum from a source enum
    """

    name: str = Field(None, description="""Target enum name""")
    populated_from: Optional[str] = Field(None, description="""Source enum name""")
    expr: Optional[str] = Field(
        None,
        description="""An expression to be evaluated on the source object to derive the target slot. Should be specified using the LinkML expression language.""",
    )
    hide: Optional[bool] = Field(None, description="""True if this is suppressed""")
    permissible_value_derivations: Optional[Dict[str, PermissibleValueDerivation]] = Field(
        default_factory=dict,
        description="""Instructions on how to derive a set of PVs in the target schema""",
    )
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[Dict[str, ElementDerivation]] = Field(default_factory=dict)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys are expressions""",
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )


class PermissibleValueDerivation(ElementDerivation):
    """
    A specification of how to derive the value of a PV from a source enum
    """

    name: str = Field(None, description="""Target permissible value text""")
    expr: Optional[str] = Field(None)
    populated_from: Optional[str] = Field(None, description="""Source permissible value""")
    hide: Optional[bool] = Field(None)
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[Dict[str, ElementDerivation]] = Field(default_factory=dict)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys are expressions""",
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )


class PrefixDerivation(ElementDerivation):
    name: str = Field(None, description="""Name of the element in the target schema""")
    copy_directives: Optional[Dict[str, CopyDirective]] = Field(default_factory=dict)
    overrides: Optional[Any] = Field(None, description="""overrides source schema slots""")
    is_a: Optional[str] = Field(None)
    mixins: Optional[Dict[str, ElementDerivation]] = Field(default_factory=dict)
    value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table that is applied directly to mappings, in order of precedence""",
    )
    expression_to_value_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys are expressions""",
    )
    expression_to_expression_mappings: Optional[Dict[str, KeyVal]] = Field(
        default_factory=dict,
        description="""A mapping table in which the keys and values are expressions""",
    )


class Inverse(ConfiguredBaseModel):
    """
    Used for back references
    """

    slot_name: Optional[str] = Field(None)
    class_name: Optional[str] = Field(None)


class KeyVal(ConfiguredBaseModel):
    key: str = Field(None)
    value: Optional[Any] = Field(None)


class CopyDirective(ConfiguredBaseModel):
    """
    Instructs a Schema Mapper in how to map to a target schema. Not used for data transformation.
    """

    element_name: str = Field(None)
    copy_all: Optional[bool] = Field(None)
    exclude_all: Optional[bool] = Field(None)
    exclude: Optional[Any] = Field(None)
    include: Optional[Any] = Field(None)
    add: Optional[Any] = Field(None)


# Update forward refs
# see https://pydantic-docs.helpmanual.io/usage/postponed_annotations/
TransformationSpecification.update_forward_refs()
ElementDerivation.update_forward_refs()
ClassDerivation.update_forward_refs()
AliasedClass.update_forward_refs()
SlotDerivation.update_forward_refs()
EnumDerivation.update_forward_refs()
PermissibleValueDerivation.update_forward_refs()
PrefixDerivation.update_forward_refs()
Inverse.update_forward_refs()
KeyVal.update_forward_refs()
CopyDirective.update_forward_refs()



from dataclasses import dataclass

from linkml_transformer.transformer.transformer import Transformer


@dataclass
class RdflibTransformer(Transformer):
    """
    A transformer that works on in-memory rdflib Graphs.

    NOT IMPLEMENTED
    """



from pathlib import Path
from typing import Union

import yaml
from linkml_runtime.processing.referencevalidator import ReferenceValidator
from linkml_runtime.utils.introspection import package_schemaview

from linkml_transformer.datamodel.transformer_model import TransformationSpecification


def load_specification(path: Union[Path, str]) -> TransformationSpecification:
    if isinstance(path, Path):
        path = str(path)
    with open(path) as f:
        obj = yaml.safe_load(f)
        # necessary to expand first
        normalizer = ReferenceValidator(
            package_schemaview("linkml_transformer.datamodel.transformer_model")
        )
        normalizer.expand_all = True
        obj = normalizer.normalize(obj)
        return TransformationSpecification(**obj)



from dataclasses import dataclass
from typing import List

from linkml_transformer.datamodel.transformer_model import (
    ElementDerivation,
    TransformationSpecification,
)


@dataclass
class TransformerSpecificationView:
    specification: TransformationSpecification = None

    def derivations(self, derivation: ElementDerivation) -> List[ElementDerivation]:
        """Return all derivations of a given derivation"""
        closure = [derivation]
        return closure



from linkml_transformer.compiler.compiler import CompiledSpecification, Compiler
from linkml_transformer.datamodel.transformer_model import TransformationSpecification


class SSSOMCompiler(Compiler):
    """
    Compiles a Transformation Specification to SSSOM.

    Note: SSSOM has less expressivity so this is expected to be highly lossy
    """

    def compile(self, specification: TransformationSpecification) -> CompiledSpecification:
        raise NotImplementedError



import logging
from collections import defaultdict
from copy import copy
from dataclasses import dataclass, field
from typing import Dict, List, Optional

from linkml_runtime import SchemaView
from linkml_runtime.linkml_model import (
    ClassDefinition,
    ClassDefinitionName,
    Element,
    SchemaDefinition,
    SlotDefinition,
)

from linkml_transformer.datamodel.transformer_model import (
    ClassDerivation,
    CopyDirective,
    TransformationSpecification,
)
from linkml_transformer.transformer.transformer import Transformer


@dataclass
class SchemaMapper:
    """
    Translates a source schema and transformation specification into a target schema.
    """

    source_schemaview: SchemaView = None

    transformer: Transformer = None

    source_to_target_class_mappings: Dict[str, List[str]] = field(
        default_factory=lambda: defaultdict(list)
    )

    def derive_schema(
        self, specification: Optional[TransformationSpecification] = None
    ) -> SchemaDefinition:
        """
        Use a transformation specification to generate a target/profile schema from a source schema.

        :param specification:
        :return:
        """
        if specification is None:
            specification = self.transformer.specification
        source_schemaview = self.source_schemaview
        source_schema = source_schemaview.schema
        target_schema = SchemaDefinition(id=source_schema.id, name=source_schema.name)
        for class_derivation in specification.class_derivations.values():
            class_definition = self._derive_class(class_derivation)
            target_schema.classes[class_definition.name] = class_definition
        for cd in target_schema.classes.values():
            self._rewire_class(cd)
        return target_schema

    def _derive_class(self, class_derivation: ClassDerivation) -> ClassDefinition:
        """
        Derive a class from a class derivation.
        """
        populated_from = class_derivation.populated_from
        if not populated_from:
            populated_from = class_derivation.name
        source_class = self.source_schemaview.get_class(populated_from)
        if source_class is None:
            logging.warning(f"No such class {populated_from}")
            target_class = ClassDefinition(name=class_derivation.name)
        else:
            target_class = copy(source_class)
            target_class.name = class_derivation.name
            target_class.slots = []
            target_class.attributes = {}
            target_class.slot_usage = {}
        for slot_derivation in class_derivation.slot_derivations.values():
            slot_definition = self._derive_slot(slot_derivation)
            target_class.attributes[slot_definition.name] = slot_definition
        self.source_to_target_class_mappings[populated_from].append(target_class.name)
        return target_class

    def _derive_slot(self, slot_derivation) -> SlotDefinition:
        """
        Derive a slot from a slot derivation.
        """
        populated_from = slot_derivation.populated_from
        if not populated_from:
            populated_from = slot_derivation.name
        source_slot = self.source_schemaview.get_slot(populated_from)
        if source_slot is None:
            target_slot = SlotDefinition(name=slot_derivation.name)
        else:
            target_slot = copy(source_slot)
            target_slot.name = slot_derivation.name
        return target_slot

    def _rewire_class(self, class_definition: ClassDefinition):
        if class_definition.is_a:
            class_definition.is_a = self._rewire_parent(class_definition, class_definition.is_a)
        mixins = [self._rewire_parent(class_definition, m) for m in class_definition.mixins]
        class_definition.mixins = [m for m in mixins if m is not None]

    def _rewire_parent(
        self, class_definition: ClassDefinition, parent: ClassDefinitionName
    ) -> Optional[str]:
        if parent in self.source_to_target_class_mappings:
            new_parents = self.source_to_target_class_mappings[parent]
            if len(new_parents) > 1:
                raise ValueError(
                    f"Cannot rewire to non-isomorphic mappings {parent} => {new_parents}"
                )
            if len(new_parents) == 1:
                return new_parents[0]
        parent_cls = self.source_schemaview.get_class(parent)
        if parent_cls.is_a:
            return self._rewire_parent(class_definition, parent_cls.is_a)
        return None

    def copy_attributes(
        self,
        target_element: Element,
        source_element: Element,
        copy_directive: CopyDirective,
    ) -> None:
        """
        Copy attributes from source to target according to a directive.

        :param target_element:
        :param source_element:
        :param copy_directive:
        :return:
        """
        for k, v in vars(source_element).items():
            included = False
            if copy_directive.include_all:
                included = True
            if k in copy_directive.include:
                included = True
            if k in copy_directive.exclude:
                included = False
            if included:
                setattr(target_element, k, v)



from abc import ABC
from dataclasses import dataclass
from typing import Iterator

from linkml_runtime import SchemaView

from linkml_transformer.datamodel.transformer_model import TransformationSpecification


@dataclass
class CompiledSpecification:
    serialization: str


@dataclass
class Compiler(ABC):
    """
    Base class for all compilers.

    A compiler will compile a transformation specification into
    an alternative representation.

    An example compiler would be a R2RML compiler.
    """

    source_schemaview: SchemaView = None
    """A view over the schema describing the source."""

    def compile(self, specification: TransformationSpecification) -> CompiledSpecification:
        """
        Transform source object into an instance of the target class.

        :param specification:
        :return:
        """
        s = ""
        for chunk in self._compile_iterator(specification):
            s += chunk
        return CompiledSpecification(serialization=s)

    def _compile_iterator(self, specification: TransformationSpecification) -> Iterator[str]:
        raise NotImplementedError
